<?php


/**
 * DB管理クラス。
 * 
 * WikiIDごとにシングルトン。
 * また、sqlite関数のラッパー。失敗したときに例外を投げる。
 */
class DataBase
{
	protected $link;	//DBへのリンク
	protected $transaction = 0;	//トランザクションのネスト数
	
	//あいまい検索用変数群
	static $fuzzypattern = array(
		'ヴァ' => 'バ',	'ヴィ' => 'ビ',	'ヴェ' => 'ベ',	'ヴォ' => 'ボ',
		'ヴ' => 'ブ',	'ヰ' => 'イ',	'ヱ' => 'エ',	'ヵ' => 'カ',
		'ァ' => 'ア',	'ィ' => 'イ',	'ゥ' => 'ウ',	'ェ' => 'エ',
		'ォ' => 'オ',	'ャ' => 'ヤ',	'ュ' => 'ユ',	'ョ' => 'ヨ');
	static $fuzzy_from;
	static $fuzzy_to;
	
	
	/**
	 * インスタンスを取得する。IDごとにシングルトン。
	 */
	static function getinstance($id = WIKIID)
	{
		static $ins = array();
		if($ins == array()){
			DataBase::$fuzzy_from = array_keys(DataBase::$fuzzypattern);
			DataBase::$fuzzy_to = array_values(DataBase::$fuzzypattern);
		}
		
		if(!isset($ins[$id])){
			$ins[$id] = new DataBase($id);
		}
		return $ins[$id];
	}
	
	
	/**
	 * コンストラクタ。
	 */
	protected function __construct($id)
	{
		$file = DATA_DIR . $id . '.db';
		$this->link = sqlite_open($file, 0666, $error);
		if($this->link == false){
			throw new MyException("DBファイル(${id}.db)を開けませんでした。\n$error");
		}
		sqlite_create_function($this->link, 'fuzzymatch', 'sqlite_udf_fuzzymatch', 2);
	}
	
	
	/**
	 * デストラクタ。
	 */
	function __destruct()
	{
		if($this->transaction > 0){
			$this->query("ROLLBACK");
		}
		if($this->link != false){
			sqlite_close($this->link);
		}
	}
	
	
	/**
	 * クエリを実行する。
	 * 
	 * @param	string	$query	SQL文。
	 * @return	resource	sqlite_query()の返値。失敗した場合は例外を投げる。
	 */
	function query($query)
	{
		$result = sqlite_unbuffered_query($this->link, $query);
		if($result == false){
			if(mb_strlen($query) > 83){
				$str = mb_substr($query, 0, 80) . '...';
			}
			else{
				$str = $query;
			}
			throw new DBException("クエリを実行できませんでした（${str}）。", $this->link);
		}
		return $result;
	}
	
	
	/**
	 * 結果セットからレコードを取得する。
	 * 
	 * @param	resource	$result	query()の返値。
	 * @return	mixed	レコードデータを含む連想配列を返す。レコードが無い場合はfalseを返す。
	 */
	function fetch($result)
	{
		$ret = sqlite_fetch_array($result);
		if(get_magic_quotes_runtime()){
			return array_map('stripslashes', $ret);
		}
		return $ret;
	}
	
	
	/**
	 * クエリパラメータ用に文字列をエスケープする。
	 * 
	 * @param	string	$str	エスケープしたい文字列。
	 * @return	string	エスケープした文字列。
	 */
	function escape($str)
	{
		//空文字列をsqlite_escape_string()に渡すと謎の3バイトが帰ってくる。
		return $str == '' ? '' : sqlite_escape_string($str);
	}
	
	
	/**
	 * 直前のクエリにより変更されたレコード数を返す。
	 * 
	 * @return	int
	 */
	function changes()
	{
		return sqlite_changes($this->link);
	}
	
	
	/**
	 * "BEGIN TRANSACTION"を発行する。
	 */
	function begin()
	{
		if($this->transaction == 0){
			$this->query("BEGIN TRANSACTION");
		}
		$this->transaction++;
	}
	
	
	/**
	 * "COMMIT"を発行する。
	 */
	function commit()
	{
		$this->transaction--;
		if($this->transaction == 0){
			$this->query("COMMIT");
		}
	}
}


/**
 * あいまい検索用UDF。
 */
function sqlite_udf_fuzzymatch($pattern, $string)
{
	//仮名は全角カタカナに、英数字・空白文字は半角に。
	$_pattern = mb_strtolower(mb_convert_kana($pattern, 'KVCas'));
	$_string = mb_strtolower(mb_convert_kana($string, 'KVCas'));
	//表記揺れの統一
	for($i = 0; $i < count(DataBase::$fuzzy_from); $i++){
		$_pattern = mb_ereg_replace(DataBase::$fuzzy_from[$i], DataBase::$fuzzy_to[$i], $_pattern);
		$_string = mb_ereg_replace(DataBase::$fuzzy_from[$i], DataBase::$fuzzy_to[$i], $_string);
	}
	//文字により削除
	$_pattern = mb_ereg_replace('[ッー・゛゜、。]', '', $_pattern);
	$_string = mb_ereg_replace('[ッー・゛゜、。]', '', $_string);
	
	return (bool)mb_ereg($_pattern, $_string);
}

?>